/* cyclops_mysql - a MySQL cheminformatics plugin using OEChem

Copyright (c) 2010 Andrew Dalke Scientific, AB, Gothenburg, Sweden

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <limits.h>
#include <ctype.h>

#include <mysql.h>
#include <oechem.h>
#include <oegraphsim.h>

using namespace std;
using namespace OEChem;
using namespace OESystem;
using namespace OEGraphSim;

#ifndef USE_MACCS_KEYS
#define USE_MACCS_KEYS 1
#endif

#if USE_MACCS_KEYS != 0 && USE_MACCS_KEYS != 1
#error USE_MACCS_KEYS must be either 0 or 1
#endif


#if USE_MACCS_KEYS == 1

struct LoadTimeInitializer {
  LoadTimeInitializer(void) {
    // There is a subtle bug in OEChem.
    // Their internal memory allocator does not do well when objects
    // are shared across different threads. This is normally not a
    // problem because MySQL uses a single thread for a search. The
    // OEMakeMACCS166FP uses a set of pre-defined SMARTS patterns.
    // Normally these could be statically allocated and instantiated
    // at load time, but they decided to defer the overhead cost (most
    // people don't need MACCS keys) and only do it on demand.
    // However, this means objects can be shared between threads,
    // which caused MySQL to crash. Horribly and sporadically.  The
    // solution is to either don't ever use MACCS keys or to bypass
    // OEChem's internal allocator and use the system one.
    // As an alternative to their own allocator, they suggest using
    // tcmalloc, which can be specified at startup time via LD_PRELOAD
    // (or equivalent for Macs)
    // You won't believe how long it took to figure this out.
    OESetMemPoolMode(OEMemPoolMode::System);
  }
};
static LoadTimeInitializer load_time_initializer;
#endif 


extern "C" {
  // oe_licensed(module)
  my_bool oe_licensed_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long oe_licensed(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void oe_licensed_deinit(UDF_INIT *initid);

  // oe_license_date(module)
  my_bool oe_license_date_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_license_date(UDF_INIT *initid, UDF_ARGS *args,
                        char *result, unsigned long *length,
                        char *is_null, char *error);
  void oe_license_date_deinit(UDF_INIT *initid);

  // oe_version(module)
  my_bool oe_version_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_version(UDF_INIT *initid, UDF_ARGS *args,
                        char *result, unsigned long *length,
                        char *is_null, char *error);
  void oe_version_deinit(UDF_INIT *initid);

  // oe_valid_smiles(smiles)
  my_bool oe_valid_smiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long oe_valid_smiles(UDF_INIT *initid, UDF_ARGS *args,
                            char *is_null, char *error);
  void oe_valid_smiles_deinit(UDF_INIT *initid);

  // oe_cansmiles(smiles)
  my_bool oe_cansmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_cansmiles(UDF_INIT *initid, UDF_ARGS *args,
                     char *result, unsigned long *length,
                     char *is_null, char *error);
  void oe_cansmiles_deinit(UDF_INIT *initid);


  // oe_isosmiles(smiles)
  my_bool oe_isosmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_isosmiles(UDF_INIT *initid, UDF_ARGS *args,
                     char *result, unsigned long *length,
                     char *is_null, char *error);
  void oe_isosmiles_deinit(UDF_INIT *initid);

  // oe_keksmiles(smiles)
  my_bool oe_keksmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_keksmiles(UDF_INIT *initid, UDF_ARGS *args,
                     char *result, unsigned long *length,
                     char *is_null, char *error);
  void oe_keksmiles_deinit(UDF_INIT *initid);

  // oe_matches(smiles, smarts)
  my_bool oe_matches_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long oe_matches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void oe_matches_deinit(UDF_INIT *initid);
  
  // oe_count_matches(smiles, smarts)
  my_bool oe_count_matches_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long oe_count_matches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void oe_count_matches_deinit(UDF_INIT *initid);

  // oe_count_umatches(smiles, smarts)
  my_bool oe_count_umatches_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long oe_count_umatches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void oe_count_umatches_deinit(UDF_INIT *initid);

  // oe_lingosim(smiles1, smiles2)
  my_bool oe_lingosim_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  double oe_lingosim(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void oe_lingosim_deinit(UDF_INIT *initid);
  

  // oe_maccs_fp(smiles)
  my_bool oe_maccs_fp_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_maccs_fp(UDF_INIT *initid, UDF_ARGS *args,
                    char *result, unsigned long *length,
                    char *is_null, char *error);
  void oe_maccs_fp_deinit(UDF_INIT *initid);

  // oe_path_fp(smiles, num_bits=4096, min_bonds=0, max_bonds=5)
  my_bool oe_path_fp_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  char *oe_path_fp(UDF_INIT *initid, UDF_ARGS *args,
                    char *result, unsigned long *length,
                    char *is_null, char *error);
  void oe_path_fp_deinit(UDF_INIT *initid);

  
  // fp_valid(fp)
  my_bool fp_valid_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long fp_valid(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void fp_valid_deinit(UDF_INIT *initid);

  // fp_contains(superstructure_fp, substructure_fp)
  my_bool fp_contains_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  long long fp_contains(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void fp_contains_deinit(UDF_INIT *initid);

  // fp_tanimoto(fp1, fp2)
  my_bool fp_tanimoto_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
  double fp_tanimoto(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
  void fp_tanimoto_deinit(UDF_INIT *initid);


}

// Parse a SMILES string and put the molecule into normalized form
// Return 1 on success, 0 on failure
static int parse_smiles(const char *smiles, int length, OEGraphMol &mol) {
  // There is a bug in OEParseSmiles when the strict flag is true
  // and the input is the empty string. I get the warning message
  //   Strict mode without previous atom.
  // followed by an illegal instruction error. I notified
  // OpenEye about this on 29 Sept 2010 and they've fixed it.
  // They say (and I agree) that the empty SMILES will not be
  // accepted under strict mode. Until everyone uses that
  // fix, I'll check for it manually.
  if (length == 0 || isspace(smiles[0]) ) {
    return 0;
  }

  string smiles_s(smiles, length);
  // input is not canonical. I like strict parsers.
  // It's open source - feel free to change that 'true' to a 'false'
  // if it makes you happy, but some of the tests will fail.
  if (!OEParseSmiles(mol, smiles_s, false, true)) {
    return 0;
  }
  OEAssignAromaticFlags(mol, OEAroModelOpenEye);
  return 1;
}
// Parse a SMARTS string. Return 1 on success, 0 on failure
static int parse_smarts(const char *smarts, int length, OESubSearch &pattern) {
  string smarts_s(smarts, length);
  return pattern.Init(smarts_s.c_str());
}

#define ERR(msg) {strcpy(message, msg); return 1;}

enum {UNKNOWN, CYCLOPS, OECHEM, OEGRAPHSIM};

static int module_name_to_int(const char *module) {
  if (strcmp(module, "oechem") == 0)
    return OECHEM;
  if (strcmp(module, "oegraphsim") == 0)
    return OEGRAPHSIM;
  if (strcmp(module, "cyclops") == 0)
    return CYCLOPS;
  return UNKNOWN;
}

static int check_module_name(UDF_ARGS *args, char *message) {
  if (args->arg_count > 1)
    ERR("too many arguments");
  
  if (args->arg_type[0] != STRING_RESULT)
    ERR("parameter 1 (module name) must be a string");

  if (args->args[0]) {
    if (module_name_to_int(args->args[0]) == UNKNOWN)
      ERR("unknown module name");
  }
  return 0;
}

// SQL: oe_licensed([module]) -> is the named module currently valid?
// If the module isn't given then return True if all licenses are available
my_bool oe_licensed_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count == 0)
    return 0;
  return check_module_name(args, message);
}

long long oe_licensed(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  if (args->arg_count == 0)
    return (OEChemIsLicensed() && OEGraphSimIsLicensed());

  switch (module_name_to_int(args->args[0])) {
  case OECHEM:
    return OEChemIsLicensed();
  case OEGRAPHSIM:
    return OEGraphSimIsLicensed();
  case CYCLOPS:
    return 1;
  default:
    *error = 1;
    return -1;
  }
}

void oe_licensed_deinit(UDF_INIT *initid) {
}

// SQL: oe_license_data(module) -> the expiration date for the cyclops,
// OEChem, or OEGraphSim license
// Date is in the format yyyy-mm-dd. Example: "2009-08-22"

my_bool oe_license_date_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count == 0)
    ERR("missing required module argument");

  return check_module_name(args, message);
}

static int check_module_name(UDF_ARGS *args) {
  return check_module_name(args);
}
char *oe_license_date(UDF_INIT *initid, UDF_ARGS *args,
                      char *result, unsigned long *length,
                      char *is_null, char *error) {
  unsigned int expdate[3];
  switch (module_name_to_int(args->args[0])) {
  case CYCLOPS:
    strcpy(result, "9999-12-31");
    *length = 10;
    return result;
  case OECHEM:
    OEChemIsLicensed(0, expdate);
    break;
  case OEGRAPHSIM:
    OEGraphSimIsLicensed(0, expdate);
    break;
  default:
    *error = 1;
    return NULL;
  }
  sprintf(result, "%04d-%02d-%02d", expdate[2], expdate[1], expdate[0]);
  *length = 10;
  return result;
}
void oe_license_date_deinit(UDF_INIT *initid) {
}

// SQL: oe_version(module) -> version information for the given module
my_bool oe_version_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count == 0)
    ERR("missing required module argument");
  return check_module_name(args, message);
}
char *oe_version(UDF_INIT *initid, UDF_ARGS *args,
                 char *result, unsigned long *length,
                 char *is_null, char *error) {
  int version;
  const char *release;

  switch (module_name_to_int(args->args[0])) {
  case CYCLOPS:
    strcpy(result, "1.0");
    *length = 3;
    return result;

  case OECHEM:
    version = OEChemGetVersion();
    release = OEChemGetRelease();
    break;

  case OEGRAPHSIM:
    version = OEGraphSimGetVersion();
    release = OEGraphSimGetRelease();
    break;

  default:
    *error = 1;
    return NULL;
  }
  int year = version / 10000;
  int month = (version % 10000) / 100;
  int day = version % 100;

  // The MySQL documentation says I'm allowed up to 255 bytes.
  // I shouldn't need anywhere near that much.
  snprintf(result, 50, "%04d-%02d-%02d %s", year, month, day, release);
  *length = strlen(result);
  return result;
}
void oe_version_deinit(UDF_INIT *initid) {
}



////////

// SQL: valid_smiles(smiles) -> is the string a value SMILES string?
my_bool oe_valid_smiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 1) {
    strcpy(message, "oe_valid_smiles takes one argument: SMILES");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "first parameter to oe_valid_smiles must be a SMILES string");
    return 1;
  }
  OEGraphMol *mol = new OEGraphMol;
  initid->ptr = reinterpret_cast<char *>(mol);
  return 0;
}

long long oe_valid_smiles(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  char *smiles = args->args[0];
  // Because we are in strict mode, work around the SMILES parsing bug
  // in OEChem when the input contains no structures. (See the comment
  // for parse_smiles.)
  if (!smiles || args->lengths[0] == 0 || isspace(smiles[0])) {
    return 0;
  }
  OEGraphMol *mol = reinterpret_cast<OEGraphMol *>(initid->ptr);
  mol->Clear();
  string smiles_s(smiles, args->lengths[0]);
  return OEParseSmiles(*mol, smiles_s, false, true);
}

void oe_valid_smiles_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OEGraphMol *>(initid->ptr);
}

///

struct OESmilesState {
  int mol_is_constant;
  OEGraphMol mol;
  string smiles_s;
};



my_bool generic_smiles_init(const char *fname, UDF_INIT *initid, UDF_ARGS *args,
                            char *message) {
  if (args->arg_count != 1) {
    sprintf(message, "%s takes one argument: SMILES", fname);
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    sprintf(message, "first parameter to %s must be a SMILES string", fname);
    return 1;
  }
  OESmilesState *state = new OESmilesState;
  char *smiles = args->args[0];
  if (smiles) {
    state->mol_is_constant = 1;
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      delete state;
      strcpy(message, "cannot parse SMILES string");
      return 1;
    }
  } else {
    state->mol_is_constant = 0;
  }
  initid->ptr = reinterpret_cast<char *>(state);
  return 0;
}

static int generic_smiles(const char *fname, OESmilesState *state,
                          UDF_INIT *initid, UDF_ARGS *args, char *error) {
  char *smiles = args->args[0];
  if (!smiles) {
    goto error;
  }
  state->mol.Clear();
  if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
    goto error;
  }
  return 0;

error:
  *error = 1;
  return 1;
}


// SQL: oe_cansmiles(smiles) -> canonical SMILES for the input
my_bool oe_cansmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smiles_init("oe_cansmiles", initid, args, message)) {
    return 1;
  }
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (state->mol_is_constant) {
    OECreateCanSmiString(state->smiles_s, state->mol);
  }
  return 0;
}

char *oe_cansmiles(UDF_INIT *initid, UDF_ARGS *args,
                   char *result, unsigned long *length,
                   char *is_null, char *error) {
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (!state->mol_is_constant) {
    if (generic_smiles("oe_cansmiles", state, initid, args, error)) {
      return NULL;
    }
    OECreateCanSmiString(state->smiles_s, state->mol);
  }
  *length = state->smiles_s.length();
  return (char *)(state->smiles_s.c_str());
}
void oe_cansmiles_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OESmilesState *>(initid->ptr);
}

// SQL: oe_isosmiles(smiles) -> isomeric SMILES for the input
my_bool oe_isosmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smiles_init("oe_isosmiles", initid, args, message)) {
    return 1;
  }
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (state->mol_is_constant) {
    OECreateIsoSmiString(state->smiles_s, state->mol);
  }
  return 0;
}

char *oe_isosmiles(UDF_INIT *initid, UDF_ARGS *args,
                   char *result, unsigned long *length,
                   char *is_null, char *error) {
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (!state->mol_is_constant) {
    if (generic_smiles("oe_isosmiles", state, initid, args, error)) {
      return NULL;
    }
    OECreateIsoSmiString(state->smiles_s, state->mol);
  }
  *length = state->smiles_s.length();
  return (char *)(state->smiles_s.c_str());
}
void oe_isosmiles_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OESmilesState *>(initid->ptr);
}

// keksmiles
// Like isosmiles but with Kekule options

const int KEKULE_ARGS = (OESMILESFlag::Kekule |
                         OESMILESFlag::Isotopes |
                         OESMILESFlag::AtomStereo |
                         OESMILESFlag::BondStereo |
                         OESMILESFlag::Canonical |
                         OESMILESFlag::AtomMaps |
                         OESMILESFlag::RGroups);

// SQL: oe_keksmiles(smiles) -> Kekule SMILES for the input
my_bool oe_keksmiles_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smiles_init("oe_cansmiles", initid, args, message)) {
    return 1;
  }
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (state->mol_is_constant) {
    OECreateSmiString(state->smiles_s, state->mol, KEKULE_ARGS);
  }
  return 0;
}


char *oe_keksmiles(UDF_INIT *initid, UDF_ARGS *args,
                   char *result, unsigned long *length,
                   char *is_null, char *error) {
  OESmilesState *state = reinterpret_cast<OESmilesState *>(initid->ptr);
  if (!state->mol_is_constant) {
    if (generic_smiles("oe_keksmiles", state, initid, args, error)) {
      return NULL;
    }
    OECreateSmiString(state->smiles_s, state->mol, KEKULE_ARGS);
  }
  *length = state->smiles_s.length();
  return (char *)(state->smiles_s.c_str());
}
void oe_keksmiles_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OESmilesState *>(initid->ptr);
}



/////

struct OEMatchState {
  int mol_is_constant;
  int pattern_is_constant;
  int result_is_constant;
  OEGraphMol mol;
  OESubSearch pattern;
  int result;
};


static int generic_smarts_match_init(const char *fname,
                                     UDF_INIT *initid, UDF_ARGS *args, char *message) {
  char *smiles, *smarts;
  if (args->arg_count != 2) {
    sprintf(message, "%s takes two arguments: SMILES, SMARTS", fname);
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    sprintf(message, "first parameter to %s must be a SMILES string", fname);
    return 1;
  }
  if (args->arg_type[1] != STRING_RESULT) {
    sprintf(message, "second parameter to %s must be a SMARTS string", fname);
    return 1;
  }
  OEMatchState *state = new OEMatchState;

  // Using a constant SMILES string?
  smiles = args->args[0];
  if (smiles) {
    state->mol_is_constant = 1;
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      sprintf(message, "%s: cannot parse SMILES string", fname);
      goto error;
    }
  } else {
    state->mol_is_constant = 0;
  }

  // Using a constant SMARTS string?
  smarts = args->args[1];
  if (smarts) {
    state->pattern_is_constant = 1;
    if (!parse_smarts(smarts, args->lengths[1], state->pattern)) {
      strcpy(message, "oe_matches: cannot parse SMARTS string");
      goto error;
    }
  } else {
    state->pattern_is_constant = 0;
  }
  // Check if the result is going to be constant
  // Can't do anything about it.
  state->result_is_constant = (state->mol_is_constant && state->pattern_is_constant);
  state->result = 0;
  initid->ptr = reinterpret_cast<char *>(state);
  return 0;

 error:
  delete state;
  return 1;
}

static int generic_smarts_match_args(const char *fname, OEMatchState *state,
                        UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  if (!state->mol_is_constant) {
    char *smiles = args->args[0];
    if (!smiles) {
      goto error;
    }
    state->mol.Clear();
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      goto error;
    }
  }
  if (!state->pattern_is_constant) {
    char *smarts = args->args[1];
    if (!smarts) {
      goto error;
    }
    if (!parse_smarts(smarts, args->lengths[1], state->pattern)) {
      goto error;
    }
  }
  return 0;
 error:
  *error = 1;
  return 1;
}



// SQL: oe_matches(smiles, smarts) -> 1 if the SMARTS matches, else 0
my_bool oe_matches_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smarts_match_init("oe_matches", initid, args, message)) {
    return 1;
  }
  // Can compute it once for the entire query?
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    state->result = state->pattern.SingleMatch(state->mol);
  }
  return 0;
}
long long oe_matches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    return state->result;
  }
  if (generic_smarts_match_args("oe_matches", state, initid, args, is_null, error)) {
    // error
    return -1;
  }
  return state->pattern.SingleMatch(state->mol);
}

void oe_matches_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OEMatchState *>(initid->ptr);
}


static int count_matches(OEGraphMol& mol, OESubSearch &pattern) {
  OEIter<OEMatchBase> match = pattern.Match(mol, 0);
  int i=0;
  while (match) {
    ++i;
    ++match;
  }
  return i;
}

static int count_umatches(OEGraphMol& mol, OESubSearch &pattern) {
  OEIter<OEMatchBase> match = pattern.Match(mol, 1);
  int i=0;
  while (match) {
    ++i;
    ++match;
  }
  return i;
}

// SQL: oe_count_matches(smiles, smarts) -> number SMARTS matches, up to 1024
my_bool oe_count_matches_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smarts_match_init("oe_count_matches", initid, args, message)) {
    return 1;
  }
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    state->result = count_matches(state->mol, state->pattern);
  }
  return 0;
}
long long oe_count_matches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    return state->result;
  }
  if (generic_smarts_match_args("oe_count_matches", state, initid, args, is_null, error)) {
    return -1;
  }
  return count_matches(state->mol, state->pattern);
}
void oe_count_matches_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OEMatchState *>(initid->ptr);
}


// SQL: oe_count_umatches(smiles, smarts) -> number unique SMARTS matches, up to 1024
my_bool oe_count_umatches_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (generic_smarts_match_init("oe_count_umatches", initid, args, message)) {
    return 1;
  }
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    state->result = count_umatches(state->mol, state->pattern);
  }
  return 0;
}
long long oe_count_umatches(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  OEMatchState *state = reinterpret_cast<OEMatchState *>(initid->ptr);
  if (state->result_is_constant) {
    return state->result;
  }
  if (generic_smarts_match_args("oe_count_umatches", state, initid, args, is_null, error)) {
    return -1;
  }
  return count_umatches(state->mol, state->pattern);
}
void oe_count_umatches_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OEMatchState *>(initid->ptr);
}

// oe_lingosim
struct OELingoSimState {
  int constant_field;  // 0 == neither, 1 == left, 2 == right
  OELingoSim sim;
};

// SQL: oe_linfosim(smiles1, smiles2) -> similarity score between 0.0 and 1.0
my_bool oe_lingosim_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 2) {
    strcpy(message, "oe_lingosim takes two (SMILES) strings");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "first parameter to oe_lingosim must be a (SMILES) string");
    return 1;
  }
  if (args->arg_type[1] != STRING_RESULT) {
    strcpy(message, "second parameter to oe_lingosim must be a (SMILES) string");
    return 1;
  }
  OELingoSimState *state = new OELingoSimState;
  initid->ptr = reinterpret_cast<char *>(state);
  char *left_s = args->args[0];
  char *right_s = args->args[1];
  if (left_s != NULL) {
    state->constant_field = 1;
    state->sim.Init(string(left_s, args->lengths[0]));
  } else if (right_s != NULL) {
    state->constant_field = 2;
    state->sim.Init(string(right_s, args->lengths[1]));
  } else {
    state->constant_field = 0;
  }
  return 0;
}
double oe_lingosim(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  OELingoSimState *state = reinterpret_cast<OELingoSimState *>(initid->ptr);
  char *left_s = args->args[0];
  char *right_s = args->args[1];
  if (!left_s || !right_s) {
    *error = 1;
    return -1;
  }
  switch (state->constant_field) {
  case 0: // neither are constant. This will be slow.
    state->sim.Init(string(left_s, args->lengths[0]));
    return state->sim.Similarity(string(right_s, args->lengths[1]));

  case 1: // left is constant
    return state->sim.Similarity(string(right_s, args->lengths[1]));

  case 2: // right is constant
    return state->sim.Similarity(string(left_s, args->lengths[0]));

  default:
    break;
  }
  fprintf(stderr, "Unknown value for lingosim constant_field: %d\n",
          state->constant_field);
  *error = 1;
  return -1;
}
void oe_lingosim_deinit(UDF_INIT *initid) {
  delete reinterpret_cast<OELingoSimState *>(initid->ptr);
}



///////////// fingerprints

struct FPState {
  int mol_is_constant;
  int num_bits;
  int min_bonds;
  int max_bonds;
  int atom_type;
  int bond_type;
  OEGraphMol mol;
  OEFingerPrint fp;
  string hex_s;
};


// Helper function to create the hex fingerprint.
// The fingerprint format is the same as OpenEye's:
//   - bytes are little endian
//   - bits in a byte are big endian
// with the modifications that
//   - fingerprint lengths must be a multiple of 8 bits
//   - there is no trailing character indicating buffer bits
//   - hex letters A-F are written in lower-case.

static const char hex_chars[] = "0123456789abcdef";

static void to_hex(int num_bits, const OEFingerPrint &fp, string &hex_s) {
  int num_bytes = (num_bits+7)/8;
  const unsigned char *data = fp.GetData();
  hex_s.resize(num_bytes*2, 'X');
  for (int i=0; i<num_bytes; i++) {
    hex_s[i*2] = hex_chars[data[i] & 0xf];
    hex_s[i*2+1] = hex_chars[data[i]>>4];
  }
}

#if USE_MACCS_KEYS == 1

// SQL: oe_maccs_fp(smiles) -> hex-encoded 166 MACCS key fingerprint
my_bool oe_maccs_fp_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 1) {
    strcpy(message, "oe_maccs_fp takes a single argument: (smiles)");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "oe_maccs_fp argument must be a SMILES string");
    return 1;
  }
  FPState *state = new FPState;

  char *smiles = args->args[0];
  if (smiles) {
    state->mol_is_constant = 1;
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      strcpy(message, "oe_maccs_fp: cannot parse SMILES string");
      goto error;
    }
    OEMakeMACCS166FP(state->fp, state->mol);
    to_hex(166, state->fp, state->hex_s);
  } else {
    state->mol_is_constant = 0;
  }
  initid->ptr = reinterpret_cast<char *>(state);
  return 0;

 error:
  delete state;
  return 1;
}

char *oe_maccs_fp(UDF_INIT *initid, UDF_ARGS *args,
               char *result, unsigned long *length,
               char *is_null, char *error) {
  FPState *state = reinterpret_cast<FPState *>(initid->ptr);
  if (state->mol_is_constant) {
    *length = 42;
    return (char *)((state->hex_s).c_str());
  }
  char *smiles = args->args[0];
  if (!smiles) {
    goto error;
  }
  state->mol.Clear();
  if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
    goto error;
  }
  OEMakeMACCS166FP(state->fp, state->mol);
  *length = 42;
  to_hex(166, state->fp, state->hex_s);
  return (char *)(state->hex_s.c_str());


 error:
  *error = 1;
  return NULL;
}

#else

my_bool oe_maccs_fp_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  sprintf(message, "cyclops_mysql was not compiled with oe_maccs_fp support");
  return 1;
}
char *oe_maccs_fp(UDF_INIT *initid, UDF_ARGS *args,
               char *result, unsigned long *length,
               char *is_null, char *error) {
  *error = 1;
  return NULL;
}
#endif

void oe_maccs_fp_deinit(UDF_INIT *initid) {
  FPState *state = reinterpret_cast<FPState *>(initid->ptr);
  delete state;
}


// If the field is available and defined, then parse it and report any range errors
static int get_oe_path_fp_args(FPState *state, UDF_ARGS *args, char *message) {
  long long val;
  if (args->arg_count > 1) {
    if (args->args[1]) {
      val = *((long long *) args->args[1]);
      if (val < 16 || val > 65536)
        ERR("argument 2 (num_bits) must be between 16 and 65536");
      if (val % 8 != 0) {
        // this is a limitation of the database implementation. I represent
        // everything through hex characters, and it's easiest to assume there
        // are two characters per byte. I could relax this in the future.
        ERR("argument 2 (num_bits) must be a multiple of 8");
      }
    } else {
      val = -1;
    }
  } else {
    val = 4096; // default num_bits is 4096
  }
  state->num_bits = val;
  
  if (args->arg_count > 2) {
    if (args->args[2]) {
      val = *((long long *) args->args[2]);
      if (val < 0 || val > INT_MAX)
        ERR("argument 3 (min_bonds) must be between 0 and maxint");
    } else {
      val = -1;
    }
  } else {
    val = 0; // default min_bonds is 0
  }
  state->min_bonds = val;

  if (args->arg_count > 3) {
    if (args->args[3]) {
      val = *((long long *) args->args[3]);
      if (val < 0 || val > INT_MAX)
        ERR("argument 4 (max_bonds) must be between 0 and maxint");
    } else {
      val = -1;
    }
  } else {
    val = 5; // default max_bonds is 0
  }
  state->max_bonds = val;

  if (state->min_bonds != -1 &&
      state->max_bonds != -1 &&
      state->min_bonds >= state->max_bonds) {
    // OpenEye issues a warning for this case:
    // Warning: Invalid min bonds > max bonds in path 0 (minbonds = maxbonds is used)
    // Since I don't want hidden modifications to the data (who will see this warning?)
    // I'll turn it into an error.
    ERR("argument 4 (max_bonds) must not be smaller than argument 3 (min_bonds)")
  }

  // I have this take integer values only, so you have to figure out
  // what the bitwise-or expression would be. I could support some
  // way to convert a "InRing|Chiral" string into the flag. But not now.
  // I want to ship this thing.
  if (args->arg_count > 4) {
    if (args->args[4]) {
      val = *((long long *) args->args[4]);
      if (val < 0 || val > 2147483647)
        ERR("argument 5 (atom_type) must be between 0 and 2**31-1");
    } else {
      val = -1;
    }
  } else {
    val = OEFPAtomType::DefaultAtom;
  }
  state->atom_type = val;


  if (args->arg_count > 5) {
    if (args->args[5]) {
      val = *((long long *) args->args[5]);
      if (val < 0 || val > 2147483647)
        ERR("argument 6 (bond_type) must be between 0 and 2**31-1");
    } else {
      val = -1;
    }
  } else {
    val = OEFPBondType::DefaultBond;
  }
  state->bond_type = val;

  return 0;
}

// SQL: oe_maccs_fp(smiles, num_bits=4096, minb=0, maxb=5, atom_type=191, bond_type=3)
//    -> hex-encoded path fingerprint, as specified
// I go through a lot of work to precompute static fields when possible.
my_bool oe_path_fp_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  const int arg_count = args->arg_count;

  // Do some basic type checking before I allocate state
  if (arg_count < 1)
    ERR("not enough arguments");
  if (arg_count > 6)
    ERR("too many arguments");
  if (arg_count == 3)
    ERR("argument 4 (max_bonds) is required if argument 3 is given");
  if (args->arg_type[0] != STRING_RESULT)
    ERR("argument 1 (smiles) must be a SMILES string");
  if (arg_count > 1 && args->arg_type[1] != INT_RESULT)
    ERR("argument 2 (num_bits) must be an integer");
  if (arg_count > 2 && args->arg_type[2] != INT_RESULT)
    ERR("argument 3 (min_bonds) must be an integer");
  if (arg_count > 3 && args->arg_type[3] != INT_RESULT)
    ERR("argument 4 (max_bonds) must be an integer");
  if (arg_count > 4 && args->arg_type[4] != INT_RESULT)
    ERR("argument 5 (atom_type) must be an integer");
  if (arg_count > 5 && args->arg_type[5] != INT_RESULT)
    ERR("argument 6 (bond_type) must be an integer");

  FPState *state = new FPState;
  char *smiles = args->args[0];
  if (smiles) {
    state->mol_is_constant = 1;
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      strcpy(message, "cannot parse SMILES string");
      goto error;
    }
  } else {
    state->mol_is_constant = 0;
  }
  if (get_oe_path_fp_args(state, args, message)) {
    goto error;
  }
  if (state->mol_is_constant &&
      state->num_bits != -1 &&
      state->min_bonds != -1 &&
      state->max_bonds != -1 &&
      state->atom_type != -1 &&
      state->bond_type != -1 ) {
    // precompute the hex fingerprint now
    OEMakePathFP(state->fp, state->mol, state->num_bits, state->min_bonds, state->max_bonds,
                 state->atom_type, state->bond_type);
    to_hex(state->num_bits, state->fp, state->hex_s);
  }
  initid->ptr = reinterpret_cast<char *>(state);
  return 0;

 error:
  delete state;
  return 1;
}

char *oe_path_fp(UDF_INIT *initid, UDF_ARGS *args,
                 char *result, unsigned long *length,
                 char *is_null, char *error) {
  FPState *state = reinterpret_cast<FPState *>(initid->ptr);
  int num_bits, min_bonds, max_bonds, atom_type, bond_type;
  long long val;
  int per_row_params = 0;
  char *smiles;

  // If the field is -1 then there was a slot for it during _init
  // but there was no value for it. It's available now. Otherwise
  // it was either a constant value or a default value.
  // NB: parameters are allowed to change on different rows
  if ((num_bits = state->num_bits) == -1) {
    per_row_params = 1;
    val = *((long long *) args->args[1]);
    if (val < 16 || val > 65536 || val % 8 != 0)
      goto error;
    num_bits = val;
  }

  if ((min_bonds = state->min_bonds) == -1) {
    per_row_params = 1;
    val = *((long long *) args->args[2]);
    if (val < 0 || val > INT_MAX)
      goto error;
    min_bonds = val;
  }

  if ((max_bonds = state->max_bonds) == -1) {
    per_row_params = 1;
    val = *((long long *) args->args[3]);
    if (val < 0 || val > INT_MAX)
      goto error;
    max_bonds = val;
  }

  if (max_bonds < min_bonds)
    goto error;  // otherwise OEChem will set minbonds = maxbonds

  if ((atom_type = state->atom_type) == -1) {
    per_row_params = 1;
    val = *((long long *) args->args[4]);
    if (val < 0 || val > 2147483647)
      goto error;
    atom_type = val;
  }

  if ((bond_type = state->bond_type) == -1) {
    per_row_params = 1;
    val = *((long long *) args->args[5]);
    if (val < 0 || val > 2147483647)
      goto error;
    bond_type = val;
  }
        
  if (state->mol_is_constant) {
    if (per_row_params) {
      // Molecule hasn't changed but args have
      OEMakePathFP(state->fp, state->mol, num_bits, min_bonds, max_bonds,
                   atom_type, bond_type);
      to_hex(num_bits, state->fp, state->hex_s);
    }
    *length = (state->hex_s).size();
    return (char *)((state->hex_s).c_str());
  } else {
    // Each row might have a different SMILES.
    smiles = args->args[0];
    if (!smiles) {
      goto error;
    }
    state->mol.Clear();
    if (!parse_smiles(smiles, args->lengths[0], state->mol)) {
      goto error;
    }
    OEMakePathFP(state->fp, state->mol, num_bits, min_bonds, max_bonds,
                 atom_type, bond_type);
    to_hex(num_bits, state->fp, state->hex_s);
  }
  *length = (state->hex_s).size();
  return (char *)((state->hex_s).c_str());

 error:
  *error = 1;
  return NULL;
}

void oe_path_fp_deinit(UDF_INIT *initid) {
  FPState *state = reinterpret_cast<FPState *>(initid->ptr);
  delete state;
}

//// functions which work with hex-encoded fingeprints

// Lookup table to map from ASCII character to hex value,
// or use a special value to indicate non-hex values.

#define BIG 16
int _char_to_value[256] = {
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  0,     1,   2,   3,   4,   5,   6,   7,   8,   9, BIG, BIG, BIG, BIG, BIG, BIG,

  BIG,  10,  11,  12,  13,  14,  15, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG,  10,  11,  12,  13,  14,  15, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,

  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,

  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
  BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG, BIG,
};

// SQL: fp_valid(fp) -> 1 if the fingerprint contains only hex characters, else 0
my_bool fp_valid_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 1) {
    strcpy(message, "fp_valid takes a single argument: hex fingerprint string");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "fp_valid single argument must be a string");
    return 1;
  }
  return 0;
}
long long fp_valid(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  unsigned char *fp = (unsigned char *) args->args[0];
  int i, len = args->lengths[0], w=0;
  if (!fp) {
    return 0;
  }
  for (i=0; i<len; i++) {
    w |= _char_to_value[fp[i]];
  }
  return (w < BIG);
}
void fp_valid_deinit(UDF_INIT *initid) {
}

// SQL: fp_contains(superstructure_fp, substructure_fp) -> 1 if the substructure
//   fingerprint is totally contained in the superstructure fingerprint
my_bool fp_contains_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 2) {
    strcpy(message, "fp_contains takes two arguments: superstructure fp, substructure fp");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "fp_contains first argument must be a fingerprint string");
    return 1;
  }
  if (args->arg_type[1] != STRING_RESULT) {
    strcpy(message, "fp_contains second argument must be a fingerprint string");
    return 1;
  }
  return 0;
}


long long fp_contains(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  unsigned char *fp1 = (unsigned char *)(args->args[0]);
  unsigned char *fp2 = (unsigned char *)(args->args[1]);
  int fp_len1, fp_len2;
  int w, w1, w2;
  if (fp1 == NULL || fp2 == NULL) {
    goto error;
  }
  fp_len1 = args->lengths[0];
  fp_len2 = args->lengths[1];
  if (fp_len1 != fp_len2) {
    goto error;
  }
  w = 0;
  for (int i=0; i<fp_len1; i++) {
    w1 = _char_to_value[fp1[i]];
    w2 = _char_to_value[fp2[i]];
    // This minimizes braches by doing the test for out-of-range
    // hex characters (which should be rare) once per 'contains'
    // test rather than once per character.
    w |= (w1|w2);
    if ((w1 & w2) != w2) {
      if (w >= BIG) {
        goto error;
      }
      return 0;
    }
  }
  if (w >= BIG) {
    goto error;
  }
  return 1;

 error:
  *error = 1;
  return -1;
}
void fp_contains_deinit(UDF_INIT *initid) {
}

// SQL: fp_tanimoto(fp1, fp2) -> the Tanimoto between the two fingerprints
my_bool fp_tanimoto_init(UDF_INIT *initid, UDF_ARGS *args, char *message) {
  if (args->arg_count != 2) {
    strcpy(message, "fp_tanimoto takes two fingerprint strings");
    return 1;
  }
  if (args->arg_type[0] != STRING_RESULT) {
    strcpy(message, "fp_tanimoto first argument must be a fingerprint string");
    return 1;
  }
  if (args->arg_type[1] != STRING_RESULT) {
    strcpy(message, "fp_tanimoto second argument must be a fingerprint string");
    return 1;
  }
  return 0;
}


int _popcount[32] = {
  0, 1, 1, 2, 1, 2, 2, 3,
  1, 2, 2, 3, 2, 3, 3, 4,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
};

double fp_tanimoto(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error) {
  int w1, w2, w=0;
  int union_count=0, intersect_count=0;
  unsigned char *fp1 = (unsigned char *)(args->args[0]);
  unsigned char *fp2 = (unsigned char *)(args->args[1]);
  int fp_len1, fp_len2;
  if (fp1 == NULL || fp2 == NULL) {
    goto error;
  }
  fp_len1 = args->lengths[0];
  fp_len2 = args->lengths[1];
  if (fp_len1 != fp_len2) {
    goto error;
  }

  for (int i=0; i<fp_len1; i++) {
    w1 = _char_to_value[fp1[i]];
    w2 = _char_to_value[fp2[i]];
    // Check for illegal characters
    w |= (w1|w2);
    // The largest possible index is (16 | 15) == 31
    union_count += _popcount[w1 | w2];
    // The largest possible index is (16 & 16) == 16
    intersect_count += _popcount[w1 & w2];
  }
  if (w >= BIG) {
    goto error;
  }
  if (union_count == 0) {
    // This is undefined. I'll define it to be 0.
    return 0.0;
  }
  return (intersect_count + 0.0) / union_count;

 error:
  *error = 1;
  return -1.0;
}

void fp_tanimoto_deinit(UDF_INIT *initid) {
}
